#  Copyright 2023 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""This module includes static variables of the agent."""

from typing import Dict

LOG_LEVEL_MAPPING: Dict[str, str] = {
    'INFO': 'INFO',
    'FAIL': 'ERROR',
    'TRACE': 'TRACE',
    'DEBUG': 'DEBUG',
    'HTML': 'INFO',
    'WARN': 'WARN',
    'ERROR': 'ERROR',
    'SKIP': 'INFO'
}
MAIN_SUITE_ID: str = 's1'
PABOT_WITHOUT_LAUNCH_ID_MSG: str = ('Pabot library is used but RP_LAUNCH_UUID was not provided. Please, '
                                    'initialize listener with the RP_LAUNCH_UUID argument.')
STATUS_MAPPING: Dict[str, str] = {
    'PASS': 'PASSED',
    'FAIL': 'FAILED',
    'NOT RUN': 'SKIPPED',
    'SKIP': 'SKIPPED'
}



#  Copyright 2024 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""This module contains functions to ease reporting to ReportPortal."""

import re
from typing import Iterable, Tuple


def replace_patterns(text: str, patterns: Iterable[Tuple[re.Pattern, str]]) -> str:
    """Replace given patterns in the text."""
    result = text
    for p, repl in patterns:
        result = p.sub(repl, result)
    return result


BARE_LINK_PATTERN = re.compile(r'\[\s*([^]|]+)]')
NAMED_LINK_PATTERN = re.compile(r'\[\s*([^]|]+)\|\s*([^]]+)]')

ROBOT_MARKUP_REPLACEMENT_PATTERS = [
    (BARE_LINK_PATTERN, r'<\1>'),
    (NAMED_LINK_PATTERN, r'[\2](\1)'),
]


def robot_markup_to_markdown(text: str) -> str:
    """Convert Robot Framework's text markup to Markdown format."""
    return replace_patterns(text, ROBOT_MARKUP_REPLACEMENT_PATTERS)



#  Copyright 2023 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""This module includes Robot Framework listener interfaces."""

import binascii
import logging
import os
import re
from functools import wraps
from mimetypes import guess_type
from typing import Optional, Dict, Union, Any
from warnings import warn

from reportportal_client.helpers import LifoQueue, is_binary, guess_content_type_from_bytes

from robotframework_reportportal.model import Keyword, Launch, Test, LogMessage, Suite
from robotframework_reportportal.service import RobotService
from robotframework_reportportal.static import MAIN_SUITE_ID, PABOT_WITHOUT_LAUNCH_ID_MSG
from robotframework_reportportal.variables import Variables

logger = logging.getLogger(__name__)
VARIABLE_PATTERN = re.compile(r'^\s*\${[^}]*}\s*=\s*')
IMAGE_PATTERN = re.compile(
    r'</td></tr><tr><td colspan="\d+"><a href="[^"]+"(?: target="_blank")?>'
    r'<img src="([^"]+)" width="\d+(?:px|pt)?"/?></a>')

DEFAULT_BINARY_FILE_TYPE = 'application/octet-stream'
TRUNCATION_SIGN = "...'"


def _unescape(binary_string: str, stop_at: int = -1):
    result = bytearray()
    join_list = list()
    join_idx = -3
    skip_next = False
    for i, b in enumerate(binary_string):
        if skip_next:
            skip_next = False
            continue
        if i < join_idx + 2:
            join_list.append(b)
            continue
        else:
            if len(join_list) > 0:
                for bb in binascii.unhexlify(''.join(join_list)):
                    result.append(bb)
                    if stop_at > 0:
                        if len(result) >= stop_at:
                            break
                join_list = list()
        if b == '\\' and binary_string[i + 1] == 'x':
            skip_next = True
            join_idx = i + 2
            continue
        for bb in b.encode('utf-8'):
            result.append(bb)
            if stop_at > 0:
                if len(result) >= stop_at:
                    break
    if len(join_list) > 0:
        for bb in binascii.unhexlify(''.join(join_list)):
            result.append(bb)
    return result


def check_rp_enabled(func):
    """Verify is RP is enabled in config."""

    @wraps(func)
    def wrap(*args, **kwargs):
        if args and isinstance(args[0], listener):
            if not args[0].service:
                return
        func(*args, **kwargs)

    return wrap


# noinspection PyPep8Naming
class listener:
    """Robot Framework listener interface for reporting to ReportPortal."""

    _items: LifoQueue
    _service: Optional[RobotService]
    _variables: Optional[Variables]
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self) -> None:
        """Initialize listener attributes."""
        self._items = LifoQueue()
        self._service = None
        self._variables = None

    def _build_msg_struct(self, message: Dict[str, Any]) -> LogMessage:
        """Check if the given message comes from our custom logger or not.

        :param message: Message passed by the Robot Framework
        """
        if isinstance(message['message'], LogMessage):
            msg = message['message']
        else:
            msg = LogMessage(message['message'])
            msg.level = message['level']
        if not msg.launch_log:
            msg.item_id = getattr(self.current_item, 'rp_item_id', None)

        message_str = msg.message
        if is_binary(message_str):
            variable_match = VARIABLE_PATTERN.search(message_str)
            if variable_match:
                # Treat as partial binary data
                msg_content = message_str[variable_match.end():]
                # remove trailing `'"...`, add `...'`
                msg.message = (message_str[variable_match.start():variable_match.end()]
                               + str(msg_content.encode('utf-8'))[:-5] + TRUNCATION_SIGN)
            else:
                # Do not log full binary data, since it's usually corrupted
                content_type = guess_content_type_from_bytes(_unescape(message_str, 128))
                msg.message = (f'Binary data of type "{content_type}" logging skipped, as it was processed as text and'
                               ' hence corrupted.')
                msg.level = 'WARN'
        elif message.get('html', 'no') == 'yes':
            image_match = IMAGE_PATTERN.match(message_str)
            if image_match:
                image_path = image_match.group(1)
                msg.message = f'Image attached: {image_path}'
                if os.path.exists(image_path):
                    image_type_by_name = guess_type(image_path)[0]
                    with open(image_path, 'rb') as fh:
                        image_data = fh.read()
                        image_type_by_data = guess_content_type_from_bytes(image_data)
                        if image_type_by_name and image_type_by_data and image_type_by_name != image_type_by_data:
                            logger.warning(
                                f'Image type mismatch: type by file name "{image_type_by_name}" '
                                f'!= type by file content "{image_type_by_data}"')
                            mime_type = DEFAULT_BINARY_FILE_TYPE
                        else:
                            mime_type = image_type_by_name or image_type_by_data or DEFAULT_BINARY_FILE_TYPE
                        msg.attachment = {
                            'name': os.path.basename(image_path),
                            'data': image_data,
                            'mime': mime_type
                        }
        return msg

    def _add_current_item(self, item: Union[Keyword, Launch, Suite, Test]) -> None:
        """Add the last item from the self._items queue."""
        self._items.put(item)

    def _remove_current_item(self) -> Union[Keyword, Launch, Suite, Test]:
        """Remove the last item from the self._items queue."""
        return self._items.get()

    @property
    def current_item(self) -> Optional[Union[Keyword, Launch, Suite, Test]]:
        """Get the last item from the self._items queue."""
        return self._items.last()

    @check_rp_enabled
    def log_message(self, message: Dict) -> None:
        """Send log message to the Report Portal.

        :param message: Message passed by the Robot Framework
        """
        msg = self._build_msg_struct(message)
        logger.debug(f'ReportPortal - Log Message: {message}')
        self.service.log(message=msg)

    @check_rp_enabled
    def log_message_with_image(self, msg: Dict, image: str):
        """Send log message to the ReportPortal.

        :param msg:   Message passed by the Robot Framework
        :param image: Path to image
        """
        mes = self._build_msg_struct(msg)
        with open(image, 'rb') as fh:
            mes.attachment = {
                'name': os.path.basename(image),
                'data': fh.read(),
                'mime': guess_type(image)[0] or DEFAULT_BINARY_FILE_TYPE
            }
        logger.debug(f'ReportPortal - Log Message with Image: {mes} {image}')
        self.service.log(message=mes)

    @property
    def parent_id(self) -> Optional[str]:
        """Get rp_item_id attribute of the current item."""
        return getattr(self.current_item, 'rp_item_id', None)

    @property
    def service(self) -> RobotService:
        """Initialize instance of the RobotService."""
        if self.variables.enabled and not self._service:
            self._service = RobotService()
            self._service.init_service(self.variables)
        return self._service

    @property
    def variables(self) -> Variables:
        """Get instance of the variables.Variables class."""
        if not self._variables:
            self._variables = Variables()
        return self._variables

    @check_rp_enabled
    def start_launch(self, attributes: Dict[str, Any], ts: Optional[Any] = None) -> None:
        """Start a new launch at the ReportPortal.

        :param attributes: Dictionary passed by the Robot Framework
        :param ts:         Timestamp(used by the ResultVisitor)
        """
        launch = Launch(self.variables.launch_name, attributes, self.variables.launch_attributes)
        launch.doc = self.variables.launch_doc or launch.doc
        if self.variables.pabot_used and not self._variables.launch_id:
            warn(PABOT_WITHOUT_LAUNCH_ID_MSG, stacklevel=2)
        logger.debug(f'ReportPortal - Start Launch: {launch.robot_attributes}')
        self.service.start_launch(
            launch=launch,
            mode=self.variables.mode,
            ts=ts,
            rerun=self.variables.rerun,
            rerun_of=self.variables.rerun_of)

    @check_rp_enabled
    def start_suite(self, name: str, attributes: Dict, ts: Optional[Any] = None) -> None:
        """Start a new test suite at the ReportPortal.

        :param name:       Test suite name
        :param attributes: Dictionary passed by the Robot Framework
        :param ts:         Timestamp(used by the ResultVisitor)
        """
        if attributes['id'] == MAIN_SUITE_ID:
            self.start_launch(attributes, ts)
            if self.variables.pabot_used:
                name = f'{name}.{self.variables.pabot_pool_id}'
            logger.debug(f'ReportPortal - Create global Suite: {attributes}')
        else:
            logger.debug(f'ReportPortal - Start Suite: {attributes}')
        suite = Suite(name, attributes)
        suite.rp_parent_item_id = self.parent_id
        suite.rp_item_id = self.service.start_suite(suite=suite, ts=ts)
        self._add_current_item(suite)

    @check_rp_enabled
    def end_suite(self, _: Optional[str], attributes: Dict, ts: Optional[Any] = None) -> None:
        """Finish started test suite at the ReportPortal.

        :param _:          Test suite name
        :param attributes: Dictionary passed by the Robot Framework
        :param ts:         Timestamp(used by the ResultVisitor)
        """
        suite = self._remove_current_item().update(attributes)
        logger.debug(f'ReportPortal - End Suite: {suite.robot_attributes}')
        self.service.finish_suite(suite=suite, ts=ts)
        if attributes['id'] == MAIN_SUITE_ID:
            launch = Launch(self.variables.launch_name, attributes, None)
            logger.debug(msg=f'ReportPortal - End Launch: {attributes}')
            self.service.finish_launch(launch=launch, ts=ts)

    @check_rp_enabled
    def start_test(self, name: str, attributes: Dict, ts: Optional[Any] = None) -> None:
        """Start a new test case at the ReportPortal.

        :param name:       Test case name
        :param attributes: Dictionary passed by the Robot Framework
        :param ts:         Timestamp(used by the ResultVisitor)
        """
        if 'source' not in attributes:
            # no 'source' parameter at this level for Robot versions < 4
            attributes = attributes.copy()
            attributes['source'] = getattr(self.current_item, 'source', None)
        test = Test(name=name, robot_attributes=attributes, test_attributes=self.variables.test_attributes)
        logger.debug(f'ReportPortal - Start Test: {attributes}')
        test.rp_parent_item_id = self.parent_id
        test.rp_item_id = self.service.start_test(test=test, ts=ts)
        self._add_current_item(test)

    @check_rp_enabled
    def end_test(self, _: Optional[str], attributes: Dict, ts: Optional[Any] = None) -> None:
        """Finish started test case at the ReportPortal.

        :param _:          Test case name
        :param attributes: Dictionary passed by the Robot Framework
        :param ts:         Timestamp(used by the ResultVisitor)
        """
        test = self.current_item.update(attributes)
        if not test.critical and test.status == 'FAIL':
            test.status = 'SKIP'
        if test.message:
            self.log_message({'message': test.message, 'level': 'DEBUG'})
        logger.debug(f'ReportPortal - End Test: {test.robot_attributes}')
        self._remove_current_item()
        self.service.finish_test(test=test, ts=ts)

    @check_rp_enabled
    def start_keyword(self, name: str, attributes: Dict, ts: Optional[Any] = None) -> None:
        """Start a new keyword(test step) at the ReportPortal.

        :param name:       Keyword name
        :param attributes: Dictionary passed by the Robot Framework
        :param ts:         Timestamp(used by the ResultVisitor)
        """
        kwd = Keyword(name=name, parent_type=self.current_item.type, robot_attributes=attributes)
        kwd.rp_parent_item_id = self.parent_id
        logger.debug(f'ReportPortal - Start Keyword: {attributes}')
        kwd.rp_item_id = self.service.start_keyword(keyword=kwd, ts=ts)
        self._add_current_item(kwd)

    @check_rp_enabled
    def end_keyword(self, _: Optional[str], attributes: Dict, ts: Optional[Any] = None) -> None:
        """Finish started keyword at the ReportPortal.

        :param _:          Keyword name
        :param attributes: Dictionary passed by the Robot Framework
        :param ts:         Timestamp(used by the ResultVisitor)
        """
        kwd = self._remove_current_item().update(attributes)
        logger.debug(f'ReportPortal - End Keyword: {kwd.robot_attributes}')
        self.service.finish_keyword(keyword=kwd, ts=ts)

    def log_file(self, log_path: str) -> None:
        """Attach HTML log file created by Robot Framework to RP launch.

        :param log_path: Path to the log file
        """
        if self.variables.attach_log:
            message = {'message': 'Execution log', 'level': 'INFO'}
            self.log_message_with_image(message, log_path)

    def report_file(self, report_path: str) -> None:
        """Attach HTML report created by Robot Framework to RP launch.

        :param report_path: Path to the report file
        """
        if self.variables.attach_report:
            message = {'message': 'Execution report', 'level': 'INFO'}
            self.log_message_with_image(message, report_path)

    def xunit_file(self, xunit_path: str) -> None:
        """Attach XUnit file created by Robot Framework to RP launch.

        :param xunit_path: Path to the XUnit file
        """
        if self.variables.attach_xunit:
            message = {'message': 'XUnit result file', 'level': 'INFO'}
            self.log_message_with_image(message, xunit_path)

    @check_rp_enabled
    def close(self) -> None:
        """Call service terminate when the whole test execution is done."""
        self.service.terminate_service()



#  Copyright (c) 2023 EPAM Systems
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License

"""This module contains model that stores Robot Framework variables."""

from os import path
from typing import Optional, Union, Dict, Tuple, Any, List
from warnings import warn

from reportportal_client import OutputType, ClientType
from reportportal_client.helpers import to_bool
from reportportal_client.logs import MAX_LOG_BATCH_PAYLOAD_SIZE
from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError

# This is a storage for the result visitor
_variables: Dict[str, Any] = {}


def get_variable(name: str, default: Optional[str] = None) -> Optional[str]:
    """Get Robot Framework variable.

    :param name:    Name of the variable
    :param default: Default value
    :return:        The value of the variable, otherwise default value
    """
    try:
        return BuiltIn().get_variable_value("${" + name + "}", default=default)
    except RobotNotRunningError:
        return _variables.get(name, default)


class Variables:
    """This class stores Robot Framework variables related to ReportPortal."""

    enabled: bool
    endpoint: Optional[str]
    launch_name: Optional[str]
    _pabot_pool_id: Optional[int]
    _pabot_used: Optional[str]
    project: Optional[str]
    api_key: Optional[str]
    attach_log: bool
    attach_report: bool
    attach_xunit: bool
    launch_attributes: List[str]
    launch_id: Optional[str]
    launch_doc: Optional[str]
    log_batch_size: Optional[int]
    mode: Optional[str]
    pool_size: Optional[int]
    rerun: bool
    rerun_of: Optional[str]
    test_attributes: List[str]
    skipped_issue: bool
    log_batch_payload_size: int
    launch_uuid_print: bool
    launch_uuid_print_output: Optional[OutputType]
    client_type: ClientType
    http_timeout: Optional[Union[Tuple[float, float], float]]

    def __init__(self) -> None:
        """Initialize instance attributes."""
        self.endpoint = get_variable('RP_ENDPOINT')
        self.launch_name = get_variable('RP_LAUNCH')
        self.project = get_variable('RP_PROJECT')

        self._pabot_pool_id = None
        self._pabot_used = None
        self.attach_log = to_bool(get_variable('RP_ATTACH_LOG', default='False'))
        self.attach_report = to_bool(get_variable('RP_ATTACH_REPORT', default='False'))
        self.attach_xunit = to_bool(get_variable('RP_ATTACH_XUNIT', default='False'))
        self.launch_attributes = get_variable('RP_LAUNCH_ATTRIBUTES', default='').split()
        self.launch_id = get_variable('RP_LAUNCH_UUID')
        self.launch_doc = get_variable('RP_LAUNCH_DOC')
        self.log_batch_size = int(get_variable(
            'RP_LOG_BATCH_SIZE', default='20'))
        self.mode = get_variable('RP_MODE')
        self.pool_size = int(get_variable('RP_MAX_POOL_SIZE', default='50'))
        self.rerun = to_bool(get_variable('RP_RERUN', default='False'))
        self.rerun_of = get_variable('RP_RERUN_OF', default=None)
        self.skipped_issue = to_bool(get_variable('RP_SKIPPED_ISSUE', default='True'))
        self.test_attributes = get_variable('RP_TEST_ATTRIBUTES', default='').split()
        self.log_batch_payload_size = int(get_variable('RP_LOG_BATCH_PAYLOAD_SIZE',
                                                       default=str(MAX_LOG_BATCH_PAYLOAD_SIZE)))
        self.launch_uuid_print = to_bool(get_variable('RP_LAUNCH_UUID_PRINT', default='False'))
        output_type = get_variable('RP_LAUNCH_UUID_PRINT_OUTPUT')
        self.launch_uuid_print_output = OutputType[output_type.upper()] if output_type else None
        client_type = get_variable('RP_CLIENT_TYPE')
        self.client_type = ClientType[client_type.upper()] if client_type else ClientType.SYNC
        connect_timeout = get_variable('RP_CONNECT_TIMEOUT')
        connect_timeout = float(connect_timeout) if connect_timeout else None

        read_timeout = get_variable('RP_READ_TIMEOUT')
        read_timeout = float(read_timeout) if read_timeout else None

        if connect_timeout is None and read_timeout is None:
            self.http_timeout = None
        elif connect_timeout is not None and read_timeout is not None:
            self.http_timeout = (connect_timeout, read_timeout)
        else:
            self.http_timeout = connect_timeout or read_timeout

        self.api_key = get_variable('RP_API_KEY')
        if not self.api_key:
            token = get_variable('RP_UUID')
            if token:
                warn(
                    message="Argument `RP_UUID` is deprecated since version 5.3.3 and will be subject for "
                            "removing in the next major version. Use `RP_API_KEY` argument instead.",
                    category=DeprecationWarning,
                    stacklevel=2
                )
                self.api_key = token
            else:
                warn(
                    message="Argument `RP_API_KEY` is `None` or empty string, that's not supposed to happen "
                            "because ReportPortal is usually requires an authorization key. Please check your"
                            " configuration.",
                    category=RuntimeWarning,
                    stacklevel=2
                )

        cond = (self.endpoint, self.launch_name, self.project, self.api_key)
        self.enabled = all(cond)
        if not self.enabled:
            warn(
                'One or required parameter is missing, ReportPortal listener will be disabled. '
                'Please check agent documentation.',
                RuntimeWarning,
                2
            )

    @property
    def pabot_pool_id(self) -> int:
        """Get pool id for the current Robot Framework executor.

        :return: Pool id for the current Robot Framework executor
        """
        if not self._pabot_pool_id:
            self._pabot_pool_id = get_variable(name='PABOTEXECUTIONPOOLID')
        return self._pabot_pool_id

    @property
    def pabot_used(self) -> Optional[str]:
        """Get status of using pabot in test execution.

        :return: Cached value of the Pabotlib URI
        """
        if not self._pabot_used:
            self._pabot_used = get_variable(name='PABOTLIBURI')
        return self._pabot_used

    @property
    def verify_ssl(self) -> Union[bool, str]:
        """Get value of the verify_ssl parameter for the client."""
        verify_ssl = get_variable('RP_VERIFY_SSL', default='True')
        if path.exists(verify_ssl):
            return verify_ssl
        return to_bool(verify_ssl)



#  Copyright 2022 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""ReportPortal logging API for test libraries with support for attachments.

Usage of this logger is similar to the standard robot.api.logger with addition
of an extra kwarg "attachment" to all logging functions.

Example
-------
import subprocess
from robotframework_reportportal import logger

class MyLibrary(object):

    def log_free_memory(self):
        logger.debug("Collecting free memory statistics!")
        logger.debug(
            "Memory consumption report",
            attachment={
                "name": "free_memory.txt",
                "data": subprocess.check_output("free -h".split()),
                "mime": "text/plain",
            },
        )
"""
from typing import Optional, Dict

from robot.api import logger

from robotframework_reportportal.model import LogMessage


def write(msg: str, level: str = 'INFO', html: bool = False, attachment: Optional[Dict[str, str]] = None,
          launch_log: bool = False) -> None:
    """Write the message to the log file using the given level.

    Valid log levels are ``TRACE``, ``DEBUG``, ``INFO`` (default since RF 2.9.1), ``WARN``,
    and ``ERROR`` (new in RF 2.9). Additionally, it is possible to use ``HTML`` pseudo log level that logs the message
    as HTML using the ``INFO`` level.

    Attachment should contain a dict with "name", "data" and "mime" values defined. See module example.

    Instead of using this method, it is generally better to use the level specific methods such as ``info`` and
    ``debug`` that have separate

    :param msg:        argument to control the message format.
    :param level:      log level
    :param html:       format or not format the message as html.
    :param attachment: a binary content to attach to the log entry
    :param launch_log: put the log entry on Launch level
    """
    log_message = LogMessage(msg)
    log_message.level = level
    log_message.attachment = attachment
    log_message.launch_log = launch_log
    logger.write(log_message, level, html)


def trace(msg: str, html: bool = False, attachment: Optional[Dict[str, str]] = None, launch_log: bool = False) -> None:
    """Write the message to the log file using the ``TRACE`` level."""
    write(msg, "TRACE", html, attachment, launch_log)


def debug(msg: str, html: bool = False, attachment: Optional[Dict[str, str]] = None, launch_log: bool = False) -> None:
    """Write the message to the log file using the ``DEBUG`` level."""
    write(msg, "DEBUG", html, attachment, launch_log)


def info(msg: str, html: bool = False, also_console: bool = False, attachment: Optional[Dict[str, str]] = None,
         launch_log: bool = False):
    """Write the message to the log file using the ``INFO`` level.

    If ``also_console`` argument is set to ``True``, the message is written both to the log file and to the console.
    """
    write(msg, "INFO", html, attachment, launch_log)
    if also_console:
        console(msg)


def warn(msg: str, html: bool = False, attachment: Optional[Dict[str, str]] = None, launch_log: bool = False) -> None:
    """Write the message to the log file using the ``WARN`` level."""
    write(msg, 'WARN', html, attachment, launch_log)


def error(msg: str, html: bool = False, attachment: Optional[Dict[str, str]] = None, launch_log: bool = False) -> None:
    """Write the message to the log file using the ``ERROR`` level."""
    write(msg, 'ERROR', html, attachment, launch_log)


def console(msg: str, newline: bool = True, stream: str = 'stdout') -> None:
    """Write the message to the console.

    If the ``newline`` argument is ``True``, a newline character is automatically added to the message.

    By default, the message is written to the standard output stream.
    Using the standard error stream is possibly by giving the ``stream`` argument value ``'stderr'``.
    """
    logger.console(msg, newline, stream)



#  Copyright 2024 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""This module contains models representing Robot Framework test items."""

import os
from typing import Any, Dict, List, Optional, Union

from robotframework_reportportal.helpers import robot_markup_to_markdown
from reportportal_client.helpers import gen_attributes

TEST_CASE_ID_SIGN = 'test_case_id:'


class Suite:
    """Class represents Robot Framework test suite."""

    robot_attributes: Union[List[str], Dict[str, Any]]
    doc: str
    end_time: str
    longname: str
    message: str
    metadata: Dict[str, str]
    name: str
    robot_id: str
    rp_item_id: Optional[str]
    rp_parent_item_id: Optional[str]
    start_time: Optional[str]
    statistics: str
    status: str
    suites: List[str]
    tests: List[str]
    total_tests: int
    type: str = 'SUITE'

    def __init__(self, name: str, robot_attributes: Dict[str, Any]):
        """Initialize required attributes.

        :param name:       Suite name
        :param robot_attributes: Suite attributes passed through the listener
        """
        self.robot_attributes = robot_attributes
        self.doc = robot_markup_to_markdown(robot_attributes['doc'])
        self.end_time = robot_attributes.get('endtime', '')
        self.longname = robot_attributes['longname']
        self.message = robot_attributes.get('message')
        self.metadata = robot_attributes['metadata']
        self.name = name
        self.robot_id = robot_attributes['id']
        self.rp_item_id = None
        self.rp_parent_item_id = None
        self.start_time = robot_attributes.get('starttime')
        self.statistics = robot_attributes.get('statistics')
        self.status = robot_attributes.get('status')
        self.suites = robot_attributes['suites']
        self.tests = robot_attributes['tests']
        self.total_tests = robot_attributes['totaltests']
        self.type = 'SUITE'

    @property
    def attributes(self) -> Optional[List[Dict[str, str]]]:
        """Get Suite attributes."""
        if self.metadata is None or not self.metadata:
            return None
        return [{'key': key, 'value': value} for key, value in self.metadata.items()]

    @property
    def source(self) -> str:
        """Return the test case source file path."""
        if self.robot_attributes.get('source') is not None:
            return os.path.relpath(self.robot_attributes['source'], os.getcwd())

    def update(self, attributes: Dict[str, Any]) -> 'Suite':
        """Update suite attributes on suite finish.

        :param attributes: Suite attributes passed through the listener
        """
        self.end_time = attributes.get('endtime', '')
        self.message = attributes.get('message')
        self.status = attributes.get('status')
        self.statistics = attributes.get('statistics')
        return self


class Launch(Suite):
    """Class represents Robot Framework test suite."""

    launch_attributes: Optional[List[Dict[str, str]]]
    type: str = 'LAUNCH'

    def __init__(self, name: str, robot_attributes: Dict[str, Any], launch_attributes: Optional[List[str]]):
        """Initialize required attributes.

        :param name:       Launch name
        :param robot_attributes: Suite attributes passed through the listener
        :param launch_attributes: Launch attributes from variables
        """
        super().__init__(name, robot_attributes)
        self.launch_attributes = gen_attributes(launch_attributes or [])
        self.type = 'LAUNCH'

    @property
    def attributes(self) -> Optional[List[Dict[str, str]]]:
        """Get Launch attributes."""
        return self.launch_attributes


class Test:
    """Class represents Robot Framework test case."""

    _critical: str
    _tags: List[str]
    robot_attributes: Dict[str, Any]
    test_attributes: Optional[List[Dict[str, str]]]
    doc: str
    end_time: str
    longname: str
    message: str
    name: str
    robot_id: str
    rp_item_id: Optional[str]
    rp_parent_item_id: Optional[str]
    start_time: str
    status: str
    template: str
    type: str = 'TEST'

    def __init__(self, name: str, robot_attributes: Dict[str, Any], test_attributes: List[str]):
        """Initialize required attributes.

        :param name:             Name of the test
        :param robot_attributes: Attributes passed through the listener
        """
        # for backward compatibility with Robot < 4.0 mark every test case
        # as critical if not set
        self._critical = robot_attributes.get('critical', 'yes')
        self._tags = robot_attributes['tags']
        self.test_attributes = gen_attributes(test_attributes)
        self.robot_attributes = robot_attributes
        self.doc = robot_markup_to_markdown(robot_attributes['doc'])
        self.end_time = robot_attributes.get('endtime', '')
        self.longname = robot_attributes['longname']
        self.message = robot_attributes.get('message')
        self.name = name
        self.robot_id = robot_attributes['id']
        self.rp_item_id = None
        self.rp_parent_item_id = None
        self.start_time = robot_attributes['starttime']
        self.status = robot_attributes.get('status')
        self.template = robot_attributes['template']
        self.type = 'TEST'

    @property
    def critical(self) -> bool:
        """Form unique value for RF 4.0+ and older versions."""
        return self._critical in ('yes', True)

    @property
    def tags(self) -> List[str]:
        """Get list of test tags excluding test_case_id."""
        return [tag for tag in self._tags if not tag.startswith(TEST_CASE_ID_SIGN)]

    @property
    def attributes(self) -> Optional[List[Dict[str, str]]]:
        """Get Test attributes."""
        return self.test_attributes + gen_attributes(self.tags)

    @property
    def source(self) -> str:
        """Return the test case source file path."""
        if self.robot_attributes['source'] is not None:
            return os.path.relpath(self.robot_attributes['source'], os.getcwd())

    @property
    def code_ref(self) -> str:
        """Return the test case code reference.

        The result line should be exactly how it appears in '.robot' file.
        """
        line_number = self.robot_attributes.get("lineno")
        if line_number is not None:
            return '{0}:{1}'.format(self.source, line_number)
        return '{0}:{1}'.format(self.source, self.name)

    @property
    def test_case_id(self) -> Optional[str]:
        """Get test case ID through the tags."""
        # use test case id from tags if specified
        for tag in self._tags:
            if tag.startswith(TEST_CASE_ID_SIGN):
                return tag.split(':')[1]
        # generate it if not
        return '{0}:{1}'.format(self.source, self.name)

    def update(self, attributes: Dict[str, Any]) -> 'Test':
        """Update test attributes on test finish.

        :param attributes: Suite attributes passed through the listener
        """
        self._tags = attributes.get('tags', self._tags)
        self.end_time = attributes.get('endtime', '')
        self.message = attributes.get('message')
        self.status = attributes.get('status')
        return self


class Keyword:
    """Class represents Robot Framework keyword."""

    robot_attributes: Dict[str, Any]
    args: List[str]
    assign: List[str]
    doc: str
    end_time: str
    keyword_name: str
    keyword_type: str
    libname: str
    name: str
    rp_item_id: Optional[str]
    rp_parent_item_id: Optional[str]
    parent_type: str
    start_time: str
    status: str
    tags: List[str]
    type: str = 'KEYWORD'

    def __init__(self, name: str, robot_attributes: Dict[str, Any], parent_type: Optional[str] = None):
        """Initialize required attributes.

        :param name:              Name of the keyword
        :param robot_attributes:  Attributes passed through the listener
        :param parent_type:       Type of the parent test item
        """
        self.robot_attributes = robot_attributes
        self.args = robot_attributes['args']
        self.assign = robot_attributes['assign']
        self.doc = robot_markup_to_markdown(robot_attributes['doc'])
        self.end_time = robot_attributes.get('endtime')
        self.keyword_name = robot_attributes['kwname']
        self.keyword_type = robot_attributes['type']
        self.libname = robot_attributes['libname']
        self.name = name
        self.rp_item_id = None
        self.rp_parent_item_id = None
        self.parent_type = parent_type
        self.start_time = robot_attributes['starttime']
        self.status = robot_attributes.get('status')
        self.tags = robot_attributes['tags']
        self.type = 'KEYWORD'

    def get_name(self) -> str:
        """Get name of the keyword suitable for ReportPortal."""
        assign = ', '.join(self.assign)
        assignment = '{0} = '.format(assign) if self.assign else ''
        arguments = ', '.join(self.args)
        full_name = f'{assignment}{self.name} ({arguments})'
        return full_name[:256]

    def get_type(self) -> str:
        """Get keyword type."""
        if self.keyword_type.lower() in ('setup', 'teardown'):
            if self.parent_type.lower() == 'keyword':
                return 'STEP'
            if self.keyword_type.lower() == 'setup':
                return 'BEFORE_{0}'.format(self.parent_type.upper())
            if self.keyword_type.lower() == 'teardown':
                return 'AFTER_{0}'.format(self.parent_type.upper())
        else:
            return 'STEP'

    def update(self, attributes: Dict[str, Any]) -> 'Keyword':
        """Update keyword attributes on keyword finish.

        :param attributes: Suite attributes passed through the listener
        """
        self.end_time = attributes.get('endtime', '')
        self.status = attributes.get('status')
        return self


class LogMessage(str):
    """Class represents Robot Framework messages."""

    attachment: Optional[Dict[str, str]]
    launch_log: bool
    item_id: Optional[str]
    level: str
    message: str

    def __init__(self, message: str):
        """Initialize required attributes."""
        self.attachment = None
        self.item_id = None
        self.level = 'INFO'
        self.launch_log = False
        self.message = message
